/*
 * MudOS mudmode connection emulator for DGD
 *
 * Based on fragments from Gurbalib and the DGD Kernel lib.
 *
 * This type of connection allows you to send LPC variables
 * between MUDs. 
 *
 * by Aidil of Way of the Force.
 */

#define DBT(X) if(find_player("aidil")) find_player("aidil")->tell(X+"\n");
# include <config.h>
# include <limits.h>
# include <privilege.h>
# include <network.h>

inherit "/kernel/lib/net/connection";
inherit "/kernel/lib/mudos/saverestore_var";

static atomic void queue_data(mixed data);

object dauser;	        /* associated user object */
int timestamp;		/* last time something was received */
string buffer;          /* receive buffer */
mixed * packets;        /* packet queue   */
int rawmode;            /* rawmode flag   */
int packet_len;         /* expected length of the current inbound packet */

/*
 * NAME:        set_rawmode()
 * DESCRIPTION: enable/disable raw tcp mode
 */
void set_rawmode(int flag) {
  rawmode = (flag != 0);
}

/*
 * NAME:        query_rawmode()
 * DESCRIPTION: return raw tcp mode status
 */
int query_rawmode() {
  return rawmode;
}

/*
 * NAME:	set_user()
 * DESCRIPTION:	set the user object
 */
void set_user(object obj) {
  if (object_name(previous_object()) == DRIVER || PRIVILEGED() || NETLIB()) {
    dauser = obj;
  }
      
  if(dauser != obj) error("Failed to set dauser!!!");
  ::set_user(dauser, "tcp");
}

/*
 * NAME:	query_player()
 * DESCRIPTION:	return the player object
 */
object query_user() {
  return dauser;
}

/*
 * NAME:	open()
 * DESCRIPTION:	start a user connection for this object
 */
void open() {
  if(previous_object() != nil) {
     error("Illegal call to open\n");
  }
  ::open(allocate(DRIVER->query_tls_size()));
  rlimits(MAXDEPTH; MAXTICKS) {
    timestamp = time();
    set_this_player(dauser);
    dauser->_F_connection(this_object());
    dauser->_F_call("open");
  }
}

/*
 * NAME:	close(varargs int dest)
 * DESCRIPTION:	close the user connection of this object
 */
static void close(varargs int dest) {
  ::close(allocate(DRIVER->query_tls_size()),dest);
}

/*
 * NAME:	force_close()
 * DESCRIPTION:	forcibly close the user connection
 */
void force_close() {
  if (!previous_object() || previous_object() == dauser || PRIVILEGED() || NETLIB() || object_name(previous_object()) == DRIVER) {
    dauser->force_close();
    dauser = nil;
    catch(destruct(this_object()));
  }
}

/*
 * NAME:	message()
 * DESCRIPTION:	convert variable to bytestream and send it to the network
 */
int message(mixed data) {
  string str;
  int size;

  if (previous_object() != dauser || data == 0) {
    return 0;
  }

  if(!rawmode) {
    size = strlen(str = save_variable(data)) + 1;

    str = "    " + str + " ";

    str[0] = (size & 0xff000000) >> 24;
    str[1] = (size & 0xff0000) >> 16;
    str[2] = (size & 0xff00) >> 8;
    str[3] = size & 0xff;

    str[strlen(str)-1] = 0;
  } else {
    if(stringp(data)) {
      str = data;
    } else {
      error("Invalid data for raw tcp mode send.\n");
    }
  }
  return ::message(str);
}

/*
 * NAME:	process_message()
 * DESCRIPTION:	process received bytestream from the network, restore variables in it
 *              and add them to the queue for the user object
 */
static void process_message(mixed *tls,string str) {
  /*
   * If we had no input, return here;
   */
  if(str == nil) return;

  /*
   * If we are in raw mode, just send whatever we get to the user object
   * and return
   */
  if(rawmode) {
    queue_data(str);
    return;
  }

  /*
   * Initialize the buffer if it doesn't exist yet.
   */
  if(!buffer) {
    buffer = "";
    packet_len = 0; /* We had no data so reset this as well */
  }

  /*
   * Decode the packets, note that sending each seperate packet to the
   * user object is handled by send_data which runs with its own
   * rlimits
   */
  rlimits (MAXDEPTH; MAXTICKS) {
    string packet;
    int size;

    /*
     * Make sure the combined input doesn't exceed our max string size.
     * This also serves to resync the connection after a while in case
     * we failed to detect the end of the previous package.
     * Buffer and expected packet size are reset here.
     */
    if( strlen( buffer ) + strlen( str ) > MAX_STRING_SIZE ) {
      buffer = "";
      packet_len = 0;
      return;
    }

    /*
     * add the input data to our receive buffer.
     */
    buffer += str;
    size = strlen(buffer);

    /* Do we have the 4 bytes for the packet size and possibly the entire packet? */
    while(size >= packet_len + 4) {
      packet_len = buffer[3] + (buffer[2] << 8) + (buffer[1] << 16) + (buffer[0] << 24);

      /* Do we really have a complete packet? */
      if(packet_len + 4 <= size) {
        /*
         * extract packet from the buffer, excluding the size and
         * trailing 0
         */
        packet = buffer[4..packet_len+2];
        /*
         * remove current packet from the buffer
         * note that we skip the trailing zero
         * from the previous packet here.
         */
        if(size > packet_len + 4) {
          buffer = buffer[packet_len+4..];
        } else {
          buffer = "";
        }
        /*
         * reset size to the current length of the buffer
         */
        size = strlen(buffer);
        /*
         * set packet length to zero, will be set to something more useful during next
         * iteration of this loop, provided there are at least 4 bytes in the buffer.
         */
        packet_len = 0; 
        /*
         * restore the variable in the packet and send it to the user object.
         */
        rlimits(MAXDEPTH;MAXTICKS) {
          queue_data(restore_variable(packet));
        } 
      } 
    }
  }
}

/*
 * NAME: receive_message()
 * DESCRIPTION: receive a packet from the network
 */
static void receive_message(string str) {
  process_message(allocate(DRIVER->query_tls_size()),str);
}

/*
 * retrieve a converted packet from the queue
 */
static atomic mixed next_packet() {
  mixed packet;

  if(!packets || !sizeof(packets)) {
    error("Queue is empty.\n");
  } else {
    packet = packets[0];
  }

  if(sizeof(packets) > 1) {
    packets = packets[1..];
  } else {
    packets = ({});
  }
  return packet;
}

/*
 * NAME:        send_data()
 * DESCRIPTION: call_out loop to send packets to the user object one at a time, and
 *              each in its own thread.
 */
void send_data() {
  mixed packet;

  if(!PRIVILEGED()) return;
  if(packets && sizeof(packets)) {
    packet = next_packet();
    catch(set_this_player(dauser));
    timestamp = time();
    /*
     * Start the call_out for the next packet before sending the
     * current one to the user, this to make sure that a runtime
     * error won't prevent sending the remaining data.
     */
    if(sizeof(packets)) call_out("send_data",0);
    dauser->receive_message(packet);
  }
}

/*
 * NAME:        queue_data()
 * DESCRIPTION: queue a variable for the user object.
 */
static atomic void queue_data(mixed data) {
  if(data) {
    if(!packets) {
      packets = ({data});
    } else {
      packets += ({data});
    }
    if(find_call_out("send_data") == -1) {
      call_out("send_data",0);
    }
  }
}

void connect(string ip, int port) {
  if(previous_object() == dauser) {
    ::connect(ip,port);
  }
}

void close_user() {
  if(previous_object() == dauser) {
    destruct_object(this_object());
  }
}

void receive_error(string message) {
  DBT("ERROR: "+message);
}
