# include <config.h>
# include <limits.h>
# include <intermud3.h>
# include <pathnames.h>
# include <loglevels.h>

#define PUBLIC_FIELD_NAME ({"old_mudlistid","old_chanlistid","telnet port","I3 TCP","I3 UDP","mudlib","base mudlib","driver","type","open","admin email","services","other data"})


private inherit "/kernel/lib/mudos/saverestore_var";

# include <router.h>

int mudlistid;
int irnmudlistid;

mapping muddata;                        /* internal mudlist */
mapping expired;                        /* expired muds, mud:changeid */
mapping queue;                          /* hold queue for pending changes */

private int DBL;
private string myname;

static mapping make_irn_mudlist_entry(string mud);
static mixed * make_mudlist_entry(string mud);
nomask int query_queue_time(string mud);
static void save_me();


static void set_mudlist_dbl(int i) {
  DBL = i;
}

static void set_myname(string str) {
  myname = str;
}

static void create() {
  string * muds;
  int i;

  if(!muddata) muddata = ([]);
  if(!expired) expired = ([]);
  if(!queue) queue = ([]);

  muds = m_indices(muddata);
  for(i=0;i<sizeof(muds);i++) {
    muddata[muds[i]]["delay"] = 0;
  }
}

static void destroying() {
  string * muds;
  int i;

  muds = m_indices(muddata);

  for(i=0;i<sizeof(muds);i++) {
    muddata[muds[i]]["delay"] = 0;
  }
}

/*
 * Find all muds that expired since the given mudlistid
 */
nomask static int filter_expired(string mud, int oldid) {
  return expired[mud] && oldid < expired[mud];
}

nomask static string * expired_since(int oldid) {
  return filter_array(m_indices(expired),"filter_expired",this_object(),oldid);
}

/*
 * Send a mudlist change concerning mud 'mud' to i3 node 'destination'
 * Note that this always uses i3v4 packets, the connection object
 * will take care of downgrading them to v3 packets when needed.
 */
nomask static int send_mudlist_delta(string mud,string destination) {
  IMUDDEBUG(DEBUG_SERVICES,"Sending mudlist entry for "+mud+" to "+destination);
  if(muddata[mud]) {
    send_from_router(
      "mudlist-delta",
      destination,
      0,
      ({
        mudlistid,
        mud,
        ({ 
          query_queue_time(mud) ? -1:muddata[mud]["delay"],
          muddata[mud]["lastip"]
        }) + make_mudlist_entry(mud)
      })
    );
  } else {
    send_from_router( "mudlist-delta", destination, 0, ({ mudlistid, mud, 0 }) );
  }
}

/*
 * Send a irn-mudlist-delta packet concerning mud 'mud' to irn router 'destination'
 */
nomask static int send_irn_mudlist_delta(string mud,string destination) {
  if(muddata[mud]) {
    if(muddata[mud]["router"] == myname) {
      IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES),"Sending irn mudlist delta (add/change) for "+mud+" to "+destination);
      send_to_node(find_irn(destination),
        ({
          "irn-mudlist-delta",
          5,
          myname,
          0,
          destination,
          0,
         irnmudlistid,
          make_irn_mudlist_entry(mud),
        })
      );
    }
  } else {
    IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES),"Sending irn mudlist delta (removal) for "+mud+" to "+destination);
    send_to_node(find_irn(destination), ({
      "irn-mudlist-delta",
      5,
      myname,
      0,
      destination,
      0,
      irnmudlistid,
      ([mud:0])
    }) );
  }
}

/*
 * Send a change to our muddata to all connected i3 nodes
 */
nomask static int broadcast_mudlist_change(string changed, varargs string * skip) {
  string * targets;
  int i;

  targets = i3_node_names();
  IMUDDEBUG(DEBUG_SERVICES,sprintf("targets for mudlist delta: %y",targets));
  if(skip) {
    targets -= skip;
    IMUDDEBUG(DEBUG_SERVICES,sprintf("targets after removal of skip array: %y",targets));
  }
  for(i=0;i<sizeof(targets);i++) {
    send_mudlist_delta(changed,targets[i]);
  }
}

/*
 * Send a chane to our muddata to all connected irn nodes
 */
nomask static void irn_broadcast_mudlist_change(string what) {
  int i;
  string * irnodes;
  mapping stuff;
  IMUDDEBUG(DEBUG_IRN | DEBUG_SERVICES, "broadcasting irn mudlist delta concerning "+what);
  stuff = make_irn_mudlist_entry(what);
  irnodes = irn_node_names();
  for(i=0;i<sizeof(irnodes);i++) {
    send_to_node(find_irn(irnodes[i]), ({"irn-mudlist-delta",5,myname,0,irnodes[i],0,irnmudlistid, stuff }) );
  }
}

/*
 * Send a complete mudlist (or all changes since oldid) to an i3 node.
 */
nomask int send_mudlist(string destination, varargs int oldid) {
  string * muds;
  int i;
  mapping result;

  muds = m_indices(muddata);

  IMUDDEBUG(DEBUG_SERVICES,"Generating mudlist for "+destination+" ("+(oldid ? oldid:0)+")");
#if 1

  for(i=0;i<sizeof(muds);i++) {
    if(!muddata[muds[i]]["changeid"] || muddata[muds[i]]["changeid"] > oldid) {
      send_mudlist_delta(muds[i],destination);
    }
  }

  if(oldid) {
    muds = expired_since(oldid);
    for(i=0;i<sizeof(muds);i++) {
      send_mudlist_delta(muds[i],destination);
    }
  }
#else
  result = ([]);
  for(i=0;i<sizeof(muds);i++) {
    if(!muddata[muds[i]]["changeid"] || muddata[muds[i]]["changeid"] > oldid) {
      result[muds[i]] = ({ 
        query_queue_time(muds[i]) ? -1:muddata[muds[i]]["delay"],
        muddata[muds[i]]["lastip"] 
      })+make_mudlist_entry(muds[i]);
    }
  }
  if(oldid) {
    muds = expired_since(oldid);
    for(i=0;i<sizeof(muds);i++) {
      result[muds[i]] = 0;
    }
  }
  IMUDDEBUG(DEBUG_SERVICES,(m_sizeof(result) ? sprintf("Sending mudlist data for %y\n",m_indices(result)):"No data to be sent\n"));
  return send_from_router( "mudlist-reply", destination, 0, ({ mudlistid, result }) );
#endif
}

/*
 * Send a complete irn mudlist to irn node 'destination'
 */
nomask static void send_irn_mudlist(string destination) {
  int i;
  string * mudnames;

  mudnames = m_indices(muddata);
  for(i=0;i<sizeof(mudnames);i++) {
    send_irn_mudlist_delta(mudnames[i],destination);
  }
}

/*
 * Our internal muddata has changed, inform the world as appropriate
 */
nomask static void mudlist_changed(string what, varargs int flag, int route) {
  mudlistid++;
  if(muddata[what]) {
    if(expired[what]) expired[what] = nil;
    muddata[what]["changeid"] = mudlistid;
    IMUDDEBUG(DEBUG_SERVICES,"Changeid for "+what+" is "+mudlistid);
  } else {
    expired[what] = mudlistid;
    IMUDDEBUG(DEBUG_SERVICES,"Added "+what+" to expired list with changeid "+mudlistid);
  }

  if(flag) {
    broadcast_mudlist_change(what,({what}));
  } else {
    broadcast_mudlist_change(what);
  }
  if(route) {
    irnmudlistid++;
    irn_broadcast_mudlist_change(what);
  }
  save_me();
}

/*
 * queue a change to the mudlist (to be processed in 5 mins if not changed
 */
nomask static void queue_mudlist_changed(string mud) {
  IMUDDEBUG(DEBUG_SERVICES,"queueing mudlist change for "+mud);
  queue[mud] = time()+300;
}

/*
 * expiration time for delayed mudlist change
 */
nomask int query_queue_time(string mud) {
  return queue[mud] != nil ? queue[mud]:0;
}

/*
 * remove delayed change from queue
 */
nomask void dequeue_mudlist_changed(string mud) {
  IMUDDEBUG(DEBUG_SERVICES,"dequeueing mudlist change for "+mud);
  queue[mud] = nil;
}

/*
 * Check the queue
 */
nomask void run_queue() {
  int i;
  string *muds;

  IMUDDEBUG(DEBUG_CORE,"running queue");
  muds = m_indices(queue);
  for(i=0;i<sizeof(muds);i++) {
    if(queue[muds[i]] < time()) {
      IMUDDEBUG(DEBUG_SERVICES,"Delayed mudlist delta concerning "+muds[i]);
      queue[muds[i]] = nil;
      mudlist_changed(muds[i],0,0);
    }
  }
}
  
/*
 * Return muddata in i3 mudlist format
 */
nomask static mixed * make_mudlist_entry(string mud) {
  if(muddata[mud]) {
    return muddata[mud]["public"][2..];
  }
}

/*
 * Return muddata in irn mudlist format
 */
nomask static mapping make_irn_mudlist_entry(string mud) {
  mapping data;

  if(!mud) error("make_irn_mudlist_entry NEEDS an argument!");
  if(!muddata[mud]) {
    return ([mud:0]);
  } else if( !mappingp(muddata[mud]) ) {
    error("muddata["+mud+"] is NOT a mapping! "+typeof(muddata[mud]));
  }

  data = muddata[mud];
  if(!data) {
    return ([mud:0]);
  } else {
    if(sizeof(data["public"]) == 11) {
      if(data["version"] != 1) {
        IMUDWARN("public data in v1 format but mud claims some other version: "+data["version"]);
      }
      data["public"] = data["public"][..9]+({""})+data["public"][10..]+({0});
    }
    return ([mud:
      (["old_mudlist_id":data["public"][0],
        "old_chanlist_id":data["public"][1],
        "player_port":data["public"][2],
        "imud_tcp_port":data["public"][3],
        "imud_udp_port":data["public"][4],
        "mudlib":data["public"][5],
        "base_mudlib":data["public"][6],
        "driver":data["public"][7],
        "mud_type":data["public"][8],
        "open_status":data["public"][9],
        "admin_email":data["public"][10],
        "services":data["public"][11],
        "other_data":data["public"][12],
        "password":data["password"],
        "connect_time":data["connect_time"],
        "disconnect_time":data["disconnect_time"],
        "name":mud,
        "ip":data["lastip"],
        "version":data["version"],
        "restart_delay":data["delay"],
        "router":data["router"]
      ])
    ]);
  }
}

nomask static void store_muddata(string mud, mapping md) {
  int pub,delay,i;

  pub = 0;
  delay = 0;

  if(!muddata[mud]) {
    IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"No muddata on "+mud);
    pub++;
  } else {
    if(sizeof(md["public"]) == 11) {
      IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"v1 mudlist data for "+mud);
      md["public"] = md["public"][0..9] + ({""})+md["public"][10..] + ({0});
    }
    for(i=0;i<sizeof(md["public"]);i++) {
      if(md["public"][i] == nil) {
        /* no nils in i3 data allowed */
        md["public"][i] = 0;
      }
      /* did anything change in the public part of the mud data? */
      if(i >= 2) {
        if(i>=sizeof(muddata[mud]["public"])) {
          IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"Field "+i+" missing in muddata[\""+mud+"\"][\"public\"]");
          pub++;
        } else if(typeof(md["public"][i]) != typeof(muddata[mud]["public"][i]) ||
          ((intp(md["public"][i]) || stringp(md["public"][i])) && md["public"][i] != muddata[mud]["public"][i]) ||
          ((mappingp(md["public"][i]) || arrayp(md["public"][i])) && save_variable(md["public"][i]) != save_variable(muddata[mud]["public"][i]))
        ) {
          IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"public change ("+(PUBLIC_FIELD_NAME[i]? PUBLIC_FIELD_NAME[i]:"unknown field "+i)+")");
          pub++;
        }
      }
    }

    if(md["lastip"] != muddata[mud]["lastip"]) {
      IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"public change (ip)");
      pub++;
    }

    if(query_queue_time(mud)) {
      if(md["delay"] != -1 && md["delay"] != muddata[mud]["delay"]) {
        IMUDDEBUG(DEBUG_SERVICES,sprintf("public change (delay was %d)",muddata[mud]["delay"]));
        IMUDDEBUG(DEBUG_SERVICES,sprintf("public change (delay is now %d)",md["delay"]));
        pub++;
        if(md["delay"] < 300) {
          delay++;
        }
      } else {
        if(md["delay"] != -1) {
          IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"Mud not online, need to send a notification to local muds");
          pub++;
          if(md["delay"] == muddata[mud]["delay"]) {
            IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"If this is the only change, requeue the notification");
            delay++;
          } else {
            IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"Notify inmediately if something changed");
          }
        } else {
          IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"Mud back online, if nothing else changed, just dequeue");
        }
      }
      dequeue_mudlist_changed(mud);
    } else if(md["delay"] != muddata[mud]["delay"]) {
      /*
       * we got a different restart_delay, check if the mud is down but
       * supposed to return soonish
       */
      if(muddata[mud]["delay"] == -1 && md["delay"] < 300) {
        IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"public change (short offline)");
        delay++;
      } else if(muddata[mud]["delay"] >= 0 && md["delay"] == -1) {
        IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"public change (online)");
      } else if(muddata[mud]["delay"] == -1) {
        IMUDDEBUG(DEBUG_SERVICES|DEBUG_IRN,"public change (long offline)");
      }
      pub++;
    }
  }

  muddata[mud] = md;

  if(pub == 1 && delay == 1) {
    queue_mudlist_changed(mud);
  } else if(pub) {
    mudlist_changed(mud,0,0);
  }
}
  
/*
 * Convert irn mudlist data to internal format and store it.
 */
nomask static void store_irn_muddata(mapping md) {
  string name;
  int i, delay, irnchange,pubchange;
  mapping data;

  name = md["name"];
  if(!name) {
    IMUDERROR("Call to store_irn_muddata but I can't find a mudname!\n");
  }

  IMUDDEBUG(DEBUG_SERVICES,"Storing irn mudlist data for "+name);

  pubchange = 0;
  delay = 0;

  if(expired[name]) {
    expired[name] = nil;
  }

  data = ([
    "delay":md["restart_delay"],
    "router":md["router"],
    "connect_time":md["connect_time"],
    "disconnect_time":md["disconnect_time"],
    "version":md["version"],
    "lastip":md["ip"],
    "password":md["password"],
    "public":({ 
      md["old_mudlist_id"],
      md["old_chanlist_id"],
      md["player_port"],
      md["imud_tcp_port"],
      md["imud_udp_port"],
      md["mudlib"],
      md["base_mudlib"],
      md["driver"],
      md["mud_type"],
      md["open_status"],
      md["admin_email"],
      md["services"],
      md["other_data"]
    })
  ]);
  store_muddata(name,data);
}

/*
 * Cleanup the mudlist
 */
void remove_mud(string str) {
  
  if(query_queue_time(str)) dequeue_mudlist_changed(str);
  muddata[str] = nil;
  mudlist_changed(str,0,1);
}

nomask static void do_cleanup() {
  int i;
  string * oldlist;
  oldlist = m_indices(muddata) + ({});
  muddata = filter_mapping(muddata,"filter_muddata",this_object());
  for(i=0;i<sizeof(oldlist);i++) {
    if(!muddata[oldlist[i]]) {
      IMUDLOG("Removed "+oldlist[i]+" from mudlist.");
      broadcast_mudlist_change(oldlist[i]);
      notify_mud_removed(oldlist[i]);
    }
  }
}

nomask mapping get_muddata(string mud) {
  return muddata[mud] ? muddata[mud][..] : nil;
}

nomask void fix_offline() {
  string * muds;
  int i;

  muds = m_indices(muddata);
  for(i=0;i<sizeof(muds);i++) {
    if(muddata[muds[i]]["delay"] == -1 && muddata[muds[i]]["connect_time"] == 0) {
      IMUDWARN("Fixed restart_delay for "+muds[i]);
      muddata[muds[i]]["delay"] = 0;
      mudlist_changed(muds[i]);
    }
  }
}

 
