/*
 * saverestore_var.c
 *
 * Emulates the MudOS save_variable and restore_variable efunctions.
 *
 * 8 January 1997       Begun           Greg Lewis
 *      # Wrote and tested initial version.  restore_variable needs to do
 *        error checking though (i.e., it works if the string it gets is
 *        right, but crashes and burns otherwise ;).
 *
 * 10 January 1997      Version 1.0     Greg Lewis
 *      # Added error checking to restore_variable, it returns RESTORE_ERROR
 *        upon errors in the restore process.
 *      # Everything seems to test ok, not that I've been all that thorough.
 *
 * TBD
 *      # Prolly needs some more commenting (doesn't most code?)
 *
 *  October 2008         Version 2.0     Bart van Leeuwen
 *      # Completely rewrote restore_variable()
 *      # It no longer returns RESTORE_ERROR upon errors, rather it returns
 *        nil. There are no other cases in which nil can be returned, this
 *        results in much better error detection (RESTORE_ERROR could also
 *        be a valid restore of an int that happens to contain that value)
 *      # objectids are not restored, at best they make no sense over a
 *        mudmode connection, at worst they are a security problem.
 */

#include <type.h>

#define GRAMMAR \
  "escaped=/\\\\/\n"+\
  "quote=/\"/\n"+\
  "floats=/[\-]?[0-9]*[.][0-9]+/\n"+\
  "arraystart=/[(][{][ ]*/\n"+\
  "arrayend=/[ ]*[}][)]/\n"+\
  "arraysep=/[ ]*,[ ]*/\n"+\
  "mapstart=/[(][[][ ]*/\n"+\
  "mapend=/[ ]*][)]/\n"+\
  "mapsep=/[ ]*:[ ]*/\n"+\
  "ints=/[\-]?[0-9]+/\n"+\
  "char=/./\n"+\
  "var: float\n"+\
  "var: int\n"+\
  "var: string\n"+\
  "var: array\n"+\
  "var: map\n"+\
  "string: quote quote ? emptystring\n"+\
  "string: quote chars quote ? concatstring\n"+\
  "array: arraystart arrayend ? mkemptyarray\n"+\
  "array: arraystart arrayelements arrayend ? mkarray\n"+\
  "map: mapstart mapend ? mkemptymap\n"+\
  "map: mapstart mapelements mapend ? mkmap\n"+\
  "arrayelements: element\n"+\
  "arrayelements: arrayelements arraysep element\n"+\
  "arrayelements: arrayelements arraysep\n"+\
  "element: int\n"+\
  "element: float\n"+\
  "element: string\n"+\
  "element: array\n"+\
  "element: map\n"+\
  "mapelements: element mapsep element\n"+\
  "mapelements: mapelements arraysep element mapsep element\n"+\
  "mapelements: mapelements arraysep\n"+\
  "chars: char\n"+\
  "chars: escaped escaped\n"+\
  "chars: escaped quote\n"+\
  "chars: arraystart\n"+\
  "chars: arrayend\n"+\
  "chars: mapstart\n"+\
  "chars: mapend\n"+\
  "chars: arraysep\n"+\
  "chars: mapsep\n"+\
  "chars: floats\n"+\
  "chars: ints\n"+\
  "chars: chars char\n"+\
  "chars: chars escaped escaped\n"+\
  "chars: chars escaped quote\n"+\
  "chars: chars arraystart\n"+\
  "chars: chars arrayend\n"+\
  "chars: chars mapstart\n"+\
  "chars: chars mapend\n"+\
  "chars: chars arraysep\n"+\
  "chars: chars mapsep\n"+\
  "chars: chars floats\n"+\
  "chars: chars ints\n"+\
  "int: ints ? cnvint\n"+\
  "float: floats ? cnvfloat\n"

int * cnvint(mixed * data) {
  int r;

  sscanf(data[0],"%d",r);
  return ({ r });
}

float * cnvfloat(mixed * data) {
  float r;

  sscanf(data[0],"%f",r);
  return ({ r });
}

object * cnvobj(mixed * data) {
  string n;
  int i;

  n = "";
  for(i=0;i<sizeof(data[0]);i++) {
    n += data[0][i];
  }
  return ({ find_object(n) });
}

string * emptystring(mixed * data) {
  return ({ "" });
}

string * concatstring(mixed * data) {
  int i;

  string r;
  r = "";

  for(i=0;i<sizeof(data);i++) {
    if(data[i] == "\\") i++;
    r += data[i];
  }

  r = implode(explode(r,"\\\""),"\"");
  r = implode(explode(r,"\\\\"),"\\");
  return ({ r[1..strlen(r)-2] });
}

mixed * mkemptyarray(mixed * data) {
  return ({ ({ }) });
}

mixed * mkarray(mixed * data) {
  int i;
  mixed * stuff;
  stuff = ({ });
  for(i=1; i<sizeof(data) && data[i] != "})"; i += 2) {
    stuff += ({ data[i] });
  }
  return ({ stuff });
}

mixed * mkemptymap(mixed * data) {
  return ({ ([ ]) });
}

mixed * mkmap(mixed * data) {
  int i;
  mapping stuff;
  mixed key,val;

  stuff = ([ ]);
  for(i=1; i<sizeof(data) && data[i] != "])"; i += 4) {
    key = data[i];
    val = data[i+2];
    stuff[key] = val;
  }
  return ({ stuff });
}

mixed parse_me(string str) {
  return parse_string(GRAMMAR,str);
}

mixed restore_variable(string str) {
  mixed * result;

  result = parse_me(str);

  if(!result || sizeof(result) < 1) {
    return nil;
  } else {
    return result[0];
  }
}

string save_variable(mixed var) {
   string       result;
   int          i, s;
   mixed        *keys, *values;
   string pref,post;

   if (!var) {
      return "0";
   }
   switch (_typeof(var)) {
      case T_INT : case T_FLOAT :
         result = ""+var;
         break;
      case T_STRING :

         if(strlen(var) > 0) {
           if(var[0..0] == "\\") {
             pref = "\\\\";
           } else if(var[0..0] == "\"") {
             pref = "\\\"";
           } else {
             pref = "";
           }

           if(var[strlen(var)-1] == '\\') {
             post = "\\\\";
           } else if(var[strlen(var)-1] == '\"') {
             post = "\\\"";
           } else {
             post = "";
           }

           result = implode(explode(var, "\\"), "\\\\");
           result = implode(explode(result, "\""), "\\\"");
           result = "\""+pref+result+post+"\"";
         } else {
           result = "\"\"";
         }
         break;
      case T_OBJECT :
         result = object_name(var);
         break;
      case T_ARRAY :
         result = "({";
         for (i = 0, s = sizeof(var); i < s; i++) {
            result += save_variable(var[i]) + ",";
         }
         result += "})";
         break;
      case T_MAPPING :
         keys = map_indices(var);
         values = map_values(var);
         result = "([";
         for (i = 0, s = map_sizeof(var); i < s; i++) {
            result += save_variable(keys[i]) + ":" + save_variable(values[i])
                      + ",";
         }
         result += "])";
         break;
      default :
         result = nil;
         break;
   }
   return result;
}

mixed testme() {
  mapping q;

  q = "/system/daemon/net/I3_router"->query_muddata();
  return restore_variable(save_variable(q));
}
